@@ -448,7 +448,7 @@ amata_next_numeric(automaton* amata, const char* prefix, char follow){
 // either a match failure or an alloc failure.
 static char*
 amata_next_string(automaton* amata, const char* prefix){
-  return amata_next_kleene(amata, prefix, '\x1b');
+  return amata_next_kleene(amata, prefix, NCKEY_ESC);
 }
 
 static inline void
@@ -1556,6 +1556,7 @@ xtversion_cb(inputctx* ictx){
     { .prefix = "contour ", .suffix = 0, .term = TERMINAL_CONTOUR, },
     { .prefix = "kitty(", .suffix = ')', .term = TERMINAL_KITTY, },
     { .prefix = "foot(", .suffix = ')', .term = TERMINAL_FOOT, },
+    { .prefix = "ghostty ", .suffix = 0, .term = TERMINAL_GHOSTTY, },
     { .prefix = "mlterm(", .suffix = ')', .term = TERMINAL_MLTERM, },
     { .prefix = "tmux ", .suffix = 0, .term = TERMINAL_TMUX, },
     { .prefix = "iTerm2 ", .suffix = 0, .term = TERMINAL_ITERM, },
@@ -1720,14 +1721,16 @@ tcap_cb(inputctx* ictx){
           ictx->initdata->qterm = TERMINAL_MLTERM;
         }else if(strcmp(key, "xterm-kitty") == 0){
           ictx->initdata->qterm = TERMINAL_KITTY;
+        }else if(strcmp(key, "xterm-ghostty") == 0){
+          ictx->initdata->qterm = TERMINAL_GHOSTTY;
         }else if(strcmp(key, "xterm-256color") == 0){
           ictx->initdata->qterm = TERMINAL_XTERM;
         }else{
-          logdebug("unknown terminal name %s", key);
+          logwarn("unknown terminal name %s", key);
         }
       }
     }else if(strcmp(val, "RGB") == 0){
-      loginfo("got rgb (%s)", s);
+      loginfo("got rgb (%s)", key);
       ictx->initdata->rgb = true;
     }else if(strcmp(val, "hpa") == 0){
       loginfo("got hpa (%s)", key);
@@ -2159,6 +2162,10 @@ prep_all_keys(inputctx* ictx){
 }
 
 // populate |buf| with any new data from the specified file descriptor |fd|.
+// bufused indicates the offset into buf at which to begin reading, and is
+// updated to reflect data read. buflen is the length of the total buffer
+// (including any skipped material). the read is nonblocking, and only one is
+// performed. this is used by the input thread.
 static void
 read_input_nblock(int fd, unsigned char* buf, size_t buflen, int *bufused,
                   unsigned* goteof){
@@ -2177,7 +2184,7 @@ read_input_nblock(int fd, unsigned char* buf, size_t buflen, int *bufused,
       if(r < 0){
         logerror("error reading from %d (%s)", fd, strerror(errno));
       }else{
-        logwarn("got EOF on %d", fd);
+        logwarn("got EOF on %d for %p", fd, goteof);
       }
       if(goteof){
         *goteof = 1;
@@ -2196,6 +2203,14 @@ ictx_independent_p(const inputctx* ictx){
   return ictx->termfd >= 0;
 }
 
+static inline void
+walk_escape_automaton(inputctx* ictx, const unsigned char* start){
+  ictx->amata.matchstart = start;
+  ictx->amata.state = ictx->amata.escapes;
+  ictx->amata.used = 1;
+  logtrace("initialized automaton to %u", ictx->amata.state);
+}
+
 // try to lex a single control sequence off of buf. return the number of bytes
 // consumed if we do so. otherwise, return the negative number of bytes
 // examined. set ictx->midescape if we're uncertain. we preserve a->used,
@@ -2222,10 +2237,7 @@ process_escape(inputctx* ictx, const unsigned char* buf, int buflen){
     // an escape always resets the trie (unless we're in the middle of an
     // ST-terminated string), as does a NULL transition.
     if(candidate == NCKEY_ESC && !ictx->amata.instring){
-      ictx->amata.matchstart = buf + ictx->amata.used - 1;
-      ictx->amata.state = ictx->amata.escapes;
-      logtrace("initialized automaton to %u", ictx->amata.state);
-      ictx->amata.used = 1;
+      walk_escape_automaton(ictx, buf + ictx->amata.used - 1);
       if(used > 1){ // we got reset; replay as input
         return -(used - 1);
       }
@@ -2236,13 +2248,14 @@ process_escape(inputctx* ictx, const unsigned char* buf, int buflen){
     }else{
       ncinput ni = {0};
       int w = walk_automaton(&ictx->amata, ictx, candidate, &ni);
-      logdebug("walk result on %u (%c): %d %u", candidate,
+      logdebug("walk result on 0x%02x (%c): %d %u", candidate,
                isprint(candidate) ? candidate : ' ', w, ictx->amata.state);
       if(w > 0){
         if(ni.id){
           load_ncinput(ictx, &ni);
         }
         ictx->amata.used = 0;
+        ictx->amata.state = 0;
         return used;
       }else if(w < 0){
         // all inspected characters are invalid; return full negative "used"
@@ -2264,7 +2277,7 @@ process_escape(inputctx* ictx, const unsigned char* buf, int buflen){
 static void
 process_escapes(inputctx* ictx, unsigned char* buf, int* bufused){
   int offset = 0;
-  while(*bufused){
+  while(*bufused > 0){
     int consumed = process_escape(ictx, buf + offset, *bufused);
     // negative |consumed| means either that we're not sure whether it's an
     // escape, or it definitely is not.
@@ -2296,7 +2309,7 @@ process_escapes(inputctx* ictx, unsigned char* buf, int* bufused){
   }
   // move any leftovers to the front; only happens if we fill output queue,
   // or ran out of input data mid-escape
-  if(*bufused){
+  if(*bufused > 0){
     ictx->amata.matchstart = buf;
     memmove(buf, buf + offset, *bufused);
   }
@@ -2375,7 +2388,7 @@ process_bulk(inputctx* ictx, unsigned char* buf, int* bufused){
     offset += consumed;
   }
   // move any leftovers to the front
-  if(*bufused){
+  if(*bufused > 0){
     memmove(buf, buf + offset, *bufused);
   }
 }
@@ -2393,7 +2406,7 @@ process_melange(inputctx* ictx, const unsigned char* buf, int* bufused){
     logdebug("input %d (%u)/%d [0x%02x] (%c)", offset, ictx->amata.used,
              *bufused, buf[offset], isprint(buf[offset]) ? buf[offset] : ' ');
     int consumed = 0;
-    if(buf[offset] == '\x1b'){
+    if(buf[offset] == NCKEY_ESC){
       consumed = process_escape(ictx, buf + offset, *bufused);
       if(consumed < 0){
         if(ictx->midescape){
@@ -2504,10 +2517,8 @@ block_on_input(inputctx* ictx, unsigned* rtfd, unsigned* rifd){
   }
   return -1;
 #else
-  int inevents = POLLIN;
-#ifdef POLLRDHUP
-  inevents |= POLLRDHUP;
-#endif
+  // do *not* use POLLRDHUP; it is not necessary, and causes trouble on BSD
+  const int inevents = POLLIN;
   struct pollfd pfds[2];
   int pfdcount = 0;
   if(!ictx->stdineof){
@@ -2668,6 +2679,11 @@ int inputready_fd(const inputctx* ictx){
 #endif
 }
 
+static void
+cleanup_mutex(void* v){
+  pthread_mutex_unlock(v);
+}
+
 static inline uint32_t
 internal_get(inputctx* ictx, const struct timespec* ts, ncinput* ni){
   uint32_t id;
@@ -2690,17 +2706,22 @@ internal_get(inputctx* ictx, const struct timespec* ts, ncinput* ni){
       }
       return NCKEY_EOF;
     }
+    int r;
+    pthread_cleanup_push(cleanup_mutex, &ictx->ilock);
     if(ts == NULL){
-      pthread_cond_wait(&ictx->icond, &ictx->ilock);
+      r = pthread_cond_wait(&ictx->icond, &ictx->ilock);
     }else{
-      int r = pthread_cond_timedwait(&ictx->icond, &ictx->ilock, ts);
+      r = pthread_cond_timedwait(&ictx->icond, &ictx->ilock, ts);
+    }
+    pthread_cleanup_pop(0);
+    if(r){
+      pthread_mutex_unlock(&ictx->ilock);
       if(r == ETIMEDOUT){
-        pthread_mutex_unlock(&ictx->ilock);
         if(ni){
           memset(ni, 0, sizeof(*ni));
         }
         return 0;
-      }else if(r < 0){
+      }else{
         inc_input_errors(ictx);
         if(ni){
           memset(ni, 0, sizeof(*ni));
@@ -2716,7 +2737,9 @@ internal_get(inputctx* ictx, const struct timespec* ts, ncinput* ni){
     if(notcurses_ucs32_to_utf8(&ni->id, 1, (unsigned char*)ni->utf8, sizeof(ni->utf8)) < 0){
       ni->utf8[0] = 0;
     }
-    if (ni->eff_text[0]==0) ni->eff_text[0]=ni->id;
+    if(ni->eff_text[0]==0){
+      ni->eff_text[0]=ni->id;
+    }
   }
   if(++ictx->iread == ictx->isize){
     ictx->iread = 0;
@@ -2728,9 +2751,11 @@ internal_get(inputctx* ictx, const struct timespec* ts, ncinput* ni){
     logtrace("draining event readiness pipe %d", ictx->ivalid);
 #ifndef __MINGW32__
     char c;
+    pthread_cleanup_push(cleanup_mutex, &ictx->ilock);
     while(read(ictx->readypipes[0], &c, sizeof(c)) == 1){
       // FIXME accelerate?
     }
+    pthread_cleanup_pop(0);
 #else
     // we ought be draining this, but it breaks everything, as we can't easily
     // do nonblocking input from a pipe in windows, augh...
